@@ -13,7 +13,8 @@ import logging
 import numpy as np
 try:
     from pydicom.dicomio import read_file
-    from pydicom.dataset import Dataset
+    from pydicom.dataset import Dataset, validate_file_meta
+    from pydicom.pixel_data_handlers.util import pixel_dtype
 except ImportError:
     from dicom import read_file
     from dicom.dataset import Dataset
@@ -31,18 +32,38 @@ if shapely_available:
 
 logger = logging.getLogger('dicompylercore.dicomparser')
 
+
 class DicomParser:
-    """Parses DICOM / DICOM RT files."""
+    """Class to parse DICOM / DICOM RT files."""
 
-    def __init__(self, dataset):
+    def __init__(self, dataset, memmap_pixel_array=False):
+        """Initialize DicomParser from a pydicom Dataset or DICOM file.
 
+        Parameters
+        ----------
+        dataset : pydicom Dataset or filename
+            pydicom dataset object or DICOM file location
+        memmap_pixel_array : bool, optional
+            Enable pixel array access via memory mapping, by default False
+
+        Raises
+        ------
+        AttributeError
+            Raised if SOPClassUID is not present in the pydicom Dataset
+        AttributeError
+            Raised if the DICOM file or pydicom Dataset cannot be read
+        """
+        self.memmap_pixel_array = memmap_pixel_array
         if isinstance(dataset, Dataset):
             self.ds = dataset
         elif isinstance(dataset, (string_types, BytesIO)):
             try:
-                self.ds = \
-                    read_file(dataset, defer_size=100, force=True)
-            except:
+                with open(dataset, "rb") as fp:
+                    self.ds = read_file(fp, defer_size=100, force=True,
+                                        stop_before_pixels=memmap_pixel_array)
+                    if memmap_pixel_array:
+                        self.offset = fp.tell() + 8
+            except Exception:
                 # Raise the error for the calling method to handle
                 raise
             else:
@@ -51,107 +72,90 @@ class DicomParser:
                 # to declare what type of file it is.
                 # If the file doesn't have a SOPClassUID,
                 # then it probably isn't DICOM.
-                if not "SOPClassUID" in self.ds:
+                if "SOPClassUID" not in self.ds:
                     raise AttributeError
         else:
             raise AttributeError
 
-######################## SOP Class and Instance Methods #######################
+        # Fix dataset file_meta if incorrect
+        try:
+            validate_file_meta(self.ds.file_meta)
+        except ValueError:
+            logger.debug('Fixing invalid File Meta for ' +
+                         str(self.ds.SOPInstanceUID))
+            self.ds.fix_meta_info()
+
+        # Remove the PixelData attribute if it is not set.
+        # i.e. RTStruct does not contain PixelData and its presence can confuse
+        # the parser
+        if "PixelData" in self.ds and self.ds.PixelData is None:
+            delattr(self.ds, 'PixelData')
+        if memmap_pixel_array:
+            self.filename = dataset
+            self.pixel_array = self.get_pixel_array
+        else:
+            if "PixelData" in self.ds:
+                self.pixel_array = self.ds.pixel_array
+
+# ====================== SOP Class and Instance Methods ======================
 
     def GetSOPClassUID(self):
         """Determine the SOP Class UID of the current file."""
+        uid = getattr(self.ds, 'SOPClassUID', None)
 
-        if (self.ds.SOPClassUID == '1.2.840.10008.5.1.4.1.1.481.2'):
+        if (uid == '1.2.840.10008.5.1.4.1.1.481.2'):
             return 'rtdose'
-        elif (self.ds.SOPClassUID == '1.2.840.10008.5.1.4.1.1.481.3'):
+        elif (uid == '1.2.840.10008.5.1.4.1.1.481.3'):
             return 'rtss'
-        elif (self.ds.SOPClassUID == '1.2.840.10008.5.1.4.1.1.481.5'):
+        elif (uid == '1.2.840.10008.5.1.4.1.1.481.5'):
             return 'rtplan'
-        elif (self.ds.SOPClassUID == '1.2.840.10008.5.1.4.1.1.2'):
+        elif (uid == '1.2.840.10008.5.1.4.1.1.2'):
             return 'ct'
         else:
             return None
 
     def GetSOPInstanceUID(self):
         """Determine the SOP Class UID of the current file."""
-
-        return self.ds.SOPInstanceUID
+        return getattr(self.ds, 'SOPInstanceUID', None)
 
     def GetStudyInfo(self):
         """Return the study information of the current file."""
-
-        study = {}
-        if 'StudyDescription' in self.ds:
-            desc = self.ds.StudyDescription
-        else:
-            desc = 'No description'
-        study['description'] = desc
-        if 'StudyDate' in self.ds:
-            date = self.ds.StudyDate
-        else:
-            date = None
-        study['date'] = date
-        if 'StudyTime' in self.ds:
-            time = self.ds.StudyTime
-        else:
-            time = None
-        study['time'] = time
-        # Don't assume that every dataset includes a study UID
-        if 'StudyInstanceUID' in self.ds:
-            study['id'] = self.ds.StudyInstanceUID
-        else:
-            study['id'] = str(random.randint(0, 65535))
-
-        return study
+        return {'description': getattr(self.ds, 'StudyDescription',
+                                       'No description'),
+                'date': getattr(self.ds, 'StudyDate', None),
+                'time': getattr(self.ds, 'StudyTime', None),
+                'id': getattr(self.ds, 'StudyInstanceUID',
+                              str(random.randint(0, 65535)))}
 
     def GetSeriesDateTime(self):
         """Return the series date/time information."""
-        dt = {}
-        if 'SeriesDate' in self.ds:
-            date = self.ds.SeriesDate
-        elif 'InstanceCreationDate' in self.ds:
-            date = self.ds.InstanceCreationDate
-        else:
-            date = None
-        dt['date'] = date
-        if 'SeriesTime' in self.ds:
-            time = self.ds.SeriesTime
-        elif 'InstanceCreationTime' in self.ds:
-            time = self.ds.InstanceCreationTime
-        else:
-            time = None
-        dt['time'] = time
+        dt = {'date': getattr(self.ds, 'SeriesDate', None),
+              'time': getattr(self.ds, 'SeriesTime', None)}
+
+        if dt['date'] is None:
+            dt['date'] = getattr(self.ds, 'InstanceCreationDate', None)
+        if dt['time'] is None:
+            dt['time'] = getattr(self.ds, 'InstanceCreationTime', None)
 
         return dt
 
     def GetSeriesInfo(self):
         """Return the series information of the current file."""
-
-        series = {}
-        if 'SeriesDescription' in self.ds:
-            desc = self.ds.SeriesDescription
-        else:
-            desc = 'No description'
-        series['description'] = desc
-        series['id'] = self.ds.SeriesInstanceUID
-        # Don't assume that every dataset includes a study UID
-        series['study'] = self.ds.SeriesInstanceUID
+        series = {'description': getattr(self.ds, 'SeriesDescription',
+                                         'No description'),
+                  'id': getattr(self.ds, 'SeriesInstanceUID', None),
+                  'study': getattr(self.ds, 'SeriesInstanceUID', None),
+                  'referenceframe': getattr(self.ds, 'FrameOfReferenceUID',
+                                            str(random.randint(0, 65535))),
+                  'modality': getattr(self.ds, 'Modality', 'OT')}
         series.update(self.GetSeriesDateTime())
-        if 'StudyInstanceUID' in self.ds:
-            series['study'] = self.ds.StudyInstanceUID
-        series['referenceframe'] = self.ds.FrameOfReferenceUID \
-            if 'FrameOfReferenceUID' in self.ds \
-            else str(random.randint(0, 65535))
-        if 'Modality' in self.ds:
-            series['modality'] = self.ds.Modality
-        else:
-            series['modality'] = 'OT'
+
+        series['study'] = getattr(self.ds, 'StudyInstanceUID', series['study'])
 
         return series
 
     def GetReferencedSeries(self):
         """Return the SOP Class UID of the referenced series."""
-
         if "ReferencedFrameOfReferenceSequence" in self.ds:
             frame = self.ds.ReferencedFrameOfReferenceSequence
             if "RTReferencedStudySequence" in frame[0]:
@@ -166,27 +170,26 @@ class DicomParser:
 
     def GetFrameOfReferenceUID(self):
         """Determine the Frame of Reference UID of the current file."""
-
         if 'FrameOfReferenceUID' in self.ds:
             # Some Files contain a Ref FoR but do not contain an FoR themselves
             if not self.ds.FrameOfReferenceUID == '':
                 return self.ds.FrameOfReferenceUID
         if 'ReferencedFrameOfReferenceSequence' in self.ds:
-            return self.ds.ReferencedFrameOfReferenceSequence[0].FrameOfReferenceUID
+            return self.ds.ReferencedFrameOfReferenceSequence[
+                0].FrameOfReferenceUID
         else:
             return ''
 
     def GetReferencedStructureSet(self):
         """Return the SOP Class UID of the referenced structure set."""
-
         if "ReferencedStructureSetSequence" in self.ds:
-            return self.ds.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID
+            return self.ds.ReferencedStructureSetSequence[
+                0].ReferencedSOPInstanceUID
         else:
             return ''
 
     def GetReferencedRTPlan(self):
         """Return the SOP Class UID of the referenced RT plan."""
-
         if "ReferencedRTPlanSequence" in self.ds:
             return self.ds.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID
         else:
@@ -194,7 +197,6 @@ class DicomParser:
 
     def GetDemographics(self):
         """Return the patient demographics from a DICOM file."""
-
         # Set up some sensible defaults for demographics
         patient = {'name': 'None',
                    'id': 'None',
@@ -223,12 +225,10 @@ class DicomParser:
 
         return patient
 
-
-############################### Image Methods #################################
+# =============================== Image Methods ===============================
 
     def GetImageData(self):
         """Return the image data from a DICOM file."""
-
         data = {}
 
         if 'ImagePositionPatient' in self.ds:
@@ -251,9 +251,35 @@ class DicomParser:
 
         return data
 
+    def GetPixelArray(self):
+        """Generate a memory mapped numpy accessor to the pixel array."""
+        if self.memmap_pixel_array is False:
+            return self.pixel_array
+        data = self.GetImageData()
+        filename = self.filename
+        dtype = pixel_dtype(self.ds)
+        offset = self.offset
+        frames = int(data['frames'])
+        shape = (int(self.GetNumberOfFrames()),
+                 data['rows'], data['columns']) if frames > 1 \
+            else (data['rows'], data['columns'])
+
+        def get_pixel_array(filename, dtype, offset, shape):
+            array = np.memmap(
+                filename,
+                dtype=dtype,
+                mode="r",
+                offset=offset,
+                shape=shape
+            )
+            yield array
+            del array
+        return list(get_pixel_array(filename, dtype, offset, shape))[0]
+
+    get_pixel_array = property(GetPixelArray)
+
     def GetImageLocation(self):
         """Calculate the location of the current image slice."""
-
         ipp = self.ds.ImagePositionPatient
         iop = self.ds.ImageOrientationPatient
 
@@ -275,7 +301,6 @@ class DicomParser:
 
     def GetImageOrientationType(self):
         """Get the orientation of the current image slice."""
-
         if 'ImageOrientationPatient' in self.ds:
             iop = np.array(self.ds.ImageOrientationPatient)
 
@@ -289,32 +314,29 @@ class DicomParser:
             ]
 
             for o in orientations:
-                if (not np.any(np.array(np.round(iop - o[1]), dtype=np.int32))):
+                if (not np.any(np.array(np.round(iop - o[1]),
+                                        dtype=np.int32))):
                     return o[0]
-        # Return N/A if the orientation was not found or could not be determined
+        # Return N/A if orientation was not found or could not be determined
         return "NA"
 
     def GetNumberOfFrames(self):
         """Return the number of frames in a DICOM image file."""
-
         frames = 1
         if 'NumberOfFrames' in self.ds:
             frames = self.ds.NumberOfFrames.real
         else:
-            try:
-                self.ds.pixel_array
-            except:
+            if "PixelData" not in self.ds:
                 return 0
             else:
-                if (self.ds.pixel_array.ndim > 2):
+                if (self.pixel_array.ndim > 2):
                     if (self.ds.SamplesPerPixel == 1) and not \
                        (self.ds.PhotometricInterpretation == 'RGB'):
-                        frames = self.ds.pixel_array.shape[0]
+                        frames = self.pixel_array.shape[0]
         return frames
 
     def GetRescaleInterceptSlope(self):
         """Return the rescale intercept and slope if present."""
-
         intercept, slope = 0, 1
         if ('RescaleIntercept' in self.ds and 'RescaleSlope' in self.ds):
             intercept = self.ds.RescaleIntercept if \
@@ -326,18 +348,36 @@ class DicomParser:
 
     def GetImage(self, window=0, level=0, size=None, background=False,
                  frames=0):
-        """Return the image from a DICOM image storage file."""
+        """Return the image from a DICOM image storage file.
 
+        Parameters
+        ----------
+        window : int, optional
+            Image window, by default 0
+        level : int, optional
+            Image level or width, by default 0
+        size : tuple, optional
+            Image size tuple in pixels, by default None
+        background : bool, optional
+            Enable a black image background, by default False
+        frames : int, optional
+            If multi-frame, use requested frame to generate image, by default 0
+
+        Returns
+        -------
+        Pillow Image
+            A Pillow Image object
+        """
         if not pil_available:
-            print("Python imaging library not available." + \
-                " Cannot generate images.")
+            print("Python imaging library not available." +
+                  " Cannot generate images.")
             return
 
-        # Return None if the Numpy pixel array cannot be accessed
+        # Return a black image if the Numpy pixel array cannot be accessed
         try:
-            self.ds.pixel_array
-        except:
-            return Image.new('RGB', size, (0, 0, 0))
+            self.pixel_array
+        except BaseException:
+            return Image.new('L', size)
 
         # Samples per pixel are > 1 & RGB format
         if (self.ds.SamplesPerPixel > 1) and \
@@ -350,7 +390,7 @@ class DicomParser:
             # Big Endian
             else:
                 im = Image.fromarray(np.rollaxis(
-                    self.ds.pixel_array.transpose(), 0, 2))
+                    self.pixel_array.transpose(), 0, 2))
 
         # Otherwise the image is monochrome
         else:
@@ -360,9 +400,9 @@ class DicomParser:
             intercept, slope = self.GetRescaleInterceptSlope()
             # Get the requested frame if multi-frame
             if (frames > 0):
-                pixel_array = self.ds.pixel_array[frames]
+                pixel_array = self.pixel_array[frames]
             else:
-                pixel_array = self.ds.pixel_array
+                pixel_array = self.pixel_array
 
             rescaled_image = pixel_array * slope + intercept
 
@@ -377,14 +417,13 @@ class DicomParser:
         if background:
             bg = Image.new('RGBA', size, (0, 0, 0, 255))
             bg.paste(im, ((size[0] - im.size[0]) // 2,
-                     (size[1] - im.size[1]) // 2))
+                          (size[1] - im.size[1]) // 2))
             return bg
 
         return im
 
     def GetDefaultImageWindowLevel(self):
         """Determine the default window/level for the DICOM image."""
-
         window, level = 0, 0
         if ('WindowWidth' in self.ds) and ('WindowCenter' in self.ds):
             if isinstance(self.ds.WindowWidth, float):
@@ -403,7 +442,7 @@ class DicomParser:
             wmin = 0
             # Rescale the slope and intercept of the image if present
             intercept, slope = self.GetRescaleInterceptSlope()
-            pixel_array = self.ds.pixel_array * slope + intercept
+            pixel_array = self.pixel_array * slope + intercept
 
             if (pixel_array.max() > wmax):
                 wmax = pixel_array.max()
@@ -416,60 +455,84 @@ class DicomParser:
         return window, level
 
     def GetLUTValue(self, data, window, level):
-        """Apply the RGB Look-Up Table for the data and window/level value."""
+        """Apply the RGB Look-Up Table for the data and window/level value.
 
-        lutvalue = util.piecewise(data,
-                                [data <= (level - 0.5 - (window - 1) / 2),
-                                 data > (level - 0.5 + (window - 1) / 2)],
-                                [0, 255, lambda data:
-                                 ((data - (level - 0.5)) / (window-1) + 0.5) *
-                                 (255 - 0)])
+        Parameters
+        ----------
+        data : numpy array
+            Pixel data array from pydicom dataset
+        window : float
+            Image window value
+        level : float
+            Image window level or width
+
+        Returns
+        -------
+        numpy array
+            Modified numpy array with RGB LUT applied
+        """
+        lutvalue = util.piecewise(
+            data,
+            [data <= (level - 0.5 - (window - 1) / 2),
+             data > (level - 0.5 + (window - 1) / 2)],
+            [0, 255, lambda data:
+             ((data - (level - 0.5)) / (window-1) + 0.5) *
+             (255 - 0)])
         # Convert the resultant array to an unsigned 8-bit array to create
         # an 8-bit grayscale LUT since the range is only from 0 to 255
         return np.array(lutvalue, dtype=np.uint8)
 
     def GetPatientToPixelLUT(self):
-        """Get the image transformation matrix from the DICOM standard Part 3
-            Section C.7.6.2.1.1"""
+        """Get image transformation matrix from the DICOM standard.
+
+        Referenced matrix can be found in Part 3 Section C.7.6.2.1.1
+        """
+        drow, dcol = self.ds.PixelSpacing
 
-        di = self.ds.PixelSpacing[0]
-        dj = self.ds.PixelSpacing[1]
         orientation = self.ds.ImageOrientationPatient
-        position = self.ds.ImagePositionPatient
+        first_x, first_y, first_z = self.ds.ImagePositionPatient
+        num_cols = self.ds.Columns
+        num_rows = self.ds.Rows
 
-        m = np.matrix(
-            [[orientation[0]*di, orientation[3]*dj, 0, position[0]],
-            [orientation[1]*di, orientation[4]*dj, 0, position[1]],
-            [orientation[2]*di, orientation[5]*dj, 0, position[2]],
-            [0, 0, 0, 1]])
+        # Determine which way X and Y real-world coords run
+        # X runs across columns if x_lut_index is 0
+        # limits to head-first/feet-first and prone/supine/decubitus
+        x_index = self.x_lut_index()
 
-        x = []
-        y = []
-        for i in range(0, self.ds.Columns):
-            imat = m * np.matrix([[i], [0], [0], [1]])
-            x.append(float(imat[0]))
-        for j in range(0, self.ds.Rows):
-            jmat = m * np.matrix([[0], [j], [0], [1]])
-            y.append(float(jmat[1]))
+        m = np.array(
+            [[orientation[0]*dcol, orientation[3]*drow, 0, first_x],
+             [orientation[1]*dcol, orientation[4]*drow, 0, first_y],
+             [orientation[2]*dcol, orientation[5]*drow, 0, first_z],
+             [0, 0, 0, 1]])
 
-        return (np.array(x), np.array(y))
+        last_xy = np.matmul(
+            m, np.array([[num_cols-1], [num_rows-1], [0], [1]])
+        )
+        last_x, last_y = last_xy[0][0], last_xy[1][0]
 
-########################## RT Structure Set Methods ###########################
+        if x_index == 0:
+            col_lut = np.linspace(first_x, last_x, num_cols)
+            row_lut = np.linspace(first_y, last_y, num_rows)
+        else:
+            col_lut = np.linspace(first_y, last_y, num_cols)
+            row_lut = np.linspace(first_x, last_x, num_rows)
+
+        return col_lut, row_lut
+
+# ========================= RT Structure Set Methods =========================
 
     def GetStructureInfo(self):
         """Return the patient demographics from a DICOM file."""
-
         structure = {}
-        structure['label'] = self.ds.StructureSetLabel
-        structure['date'] = self.ds.StructureSetDate
-        structure['time'] = self.ds.StructureSetTime
+        structure['label'] = getattr(self.ds, 'StructureSetLabel', '')
+        structure['date'] = getattr(self.ds, 'StructureSetDate', '')
+        structure['time'] = getattr(self.ds, 'StructureSetTime', '')
         structure['numcontours'] = len(self.ds.ROIContourSequence)
 
         return structure
 
     def GetStructures(self):
-        """Returns a dictionary of structures (ROIs)."""
-
+        """Return a dictionary of structures (ROIs)."""
         structures = {}
 
         # Determine whether this is RT Structure Set file
@@ -517,7 +580,7 @@ class DicomParser:
                         structures[number]['color'] = \
                             np.array(color, dtype=int)
                     # Otherwise fail and fallback on the random color
-                    except:
+                    except Exception:
                         logger.debug(
                             "Unable to decode display color for ROI #%s",
                             str(number))
@@ -531,8 +594,18 @@ class DicomParser:
         return structures
 
     def GetStructureCoordinates(self, roi_number):
-        """Get the list of coordinates for each plane of the structure."""
+        """Get the list of coordinates for each plane of the structure.
 
+        Parameters
+        ----------
+        roi_number : integer
+            ROI number used to index structure from RT Struct
+
+        Returns
+        -------
+        dict
+            Dict of structure coordinates sorted by slice position (z)
+        """
         planes = {}
         # The coordinate data of each ROI is stored within ROIContourSequence
         if 'ROIContourSequence' in self.ds:
@@ -547,6 +620,10 @@ class DicomParser:
                             # Determine all the plane properties
                             plane['type'] = c.ContourGeometricType
                             plane['num_points'] = int(c.NumberOfContourPoints)
+                            # Since DICOM RT Structure Set (C.8.8.6) specifies
+                            # that a ContourData is stored as an flattened list
+                            # of xyz triples, convert it to a unflattened list
+                            # for easier parsing
                             plane['data'] = \
                                 self.GetContourPoints(c.ContourData)
 
@@ -560,14 +637,34 @@ class DicomParser:
         return planes
 
     def GetContourPoints(self, array):
-        """Parses an array of xyz points & returns a array of point dicts."""
+        """Unflatten a flattened list of xyz point triples.
 
+        Parameters
+        ----------
+        array : list
+            Flattened list of xyz point triples
+
+        Returns
+        -------
+        list
+            Unflattened list of xyz point triples
+        """
         n = 3
         return [array[i:i+n] for i in range(0, len(array), n)]
 
     def CalculatePlaneThickness(self, planesDict):
-        """Calculates the plane thickness for each structure."""
+        """Calculate the plane thickness for each structure.
 
+        Parameters
+        ----------
+        planesDict : dict
+            Output from GetStructureCoordinates
+
+        Returns
+        -------
+        float
+            Thickness of the structure in mm
+        """
         planes = []
 
         # Iterate over each plane in the structure
@@ -590,22 +687,26 @@ class DicomParser:
         return thickness
 
     def CalculateStructureVolume(self, coords, thickness):
-        """Calculates the volume of the given structure.
+        """Calculate the volume of the given structure.
 
         Parameters
         ----------
         coords : dict
             Coordinates of each plane of the structure
         thickness : float
-            Thickness of the structure
-        """
+            Thickness of the structure in mm
 
+        Returns
+        -------
+        float
+            Volume of structure in cm3
+        """
         if not shapely_available:
             print("Shapely library not available." +
                   " Please install to calculate.")
             return 0
 
-        class Within(Polygon):
+        class Within:
             def __init__(self, o):
                 self.o = o
 
@@ -645,25 +746,23 @@ class DicomParser:
         vol = s * thickness / 1000
         return vol
 
-############################## RT Dose Methods ###############################
+# ============================== RT Dose Methods ==============================
 
     def HasDVHs(self):
-        """Returns whether dose-volume histograms (DVHs) exist."""
-
-        if not "DVHSequence" in self.ds:
+        """Return whether dose-volume histograms (DVHs) exist."""
+        if "DVHSequence" not in self.ds:
             return False
         else:
             return True
 
     def GetDVHs(self):
-        """Returns cumulative dose-volume histograms (DVHs)."""
-
+        """Return cumulative dose-volume histograms (DVHs)."""
         self.dvhs = {}
 
         if self.HasDVHs():
             for item in self.ds.DVHSequence:
                 # Make sure that the DVH has a referenced structure / ROI
-                if not 'DVHReferencedROISequence' in item:
+                if 'DVHReferencedROISequence' not in item:
                     continue
                 number = item.DVHReferencedROISequence[0].ReferencedROINumber
                 logger.debug("Found DVH for ROI #%s", str(number))
@@ -672,34 +771,43 @@ class DicomParser:
         return self.dvhs
 
     def GetDoseGrid(self, z=0, threshold=0.5):
-        """
-        Return the 2d dose grid for the given slice position (mm).
+        """Return the 2d dose grid for the given slice position (mm).
 
-        :param z:           Slice position in mm.
-        :param threshold:   Threshold in mm to determine the max difference
-                            from z to the closest dose slice without
-                            using interpolation.
-        :return:            An numpy 2d array of dose points.
-        """
+        Parameters
+        ----------
+        z : int, optional
+            Slice position in mm, by default 0
+        threshold : float, optional
+            Threshold in mm to determine the max difference from z
+            to the closest dose slice without using interpolation,
+            by default 0.5
 
+        Returns
+        -------
+        np.array
+            An numpy 2d array of dose points
+        """
         # If this is a multi-frame dose pixel array,
         # determine the offset for each frame
         if 'GridFrameOffsetVector' in self.ds:
+            pixel_array = self.GetPixelArray()
             z = float(z)
             # Get the initial dose grid position (z) in patient coordinates
-            imagepatpos = self.ds.ImagePositionPatient[2]
-            orientation = self.ds.ImageOrientationPatient[0]
+            ipp = self.ds.ImagePositionPatient
+            gfov = self.ds.GridFrameOffsetVector
             # Add the position to the offset vector to determine the
             # z coordinate of each dose plane
-            planes = orientation * np.array(self.ds.GridFrameOffsetVector) + \
-                imagepatpos
+
+            z_sign = 1 if self.is_head_first_orientation() else -1
+
+            planes = (z_sign * np.array(gfov)) + ipp[2]
             frame = -1
             # Check to see if the requested plane exists in the array
             if (np.amin(np.fabs(planes - z)) < threshold):
                 frame = np.argmin(np.fabs(planes - z))
             # Return the requested dose plane, since it was found
             if not (frame == -1):
-                return self.ds.pixel_array[frame]
+                return pixel_array[frame]
             # Check if the requested plane is within the dose grid boundaries
             elif ((z < np.amin(planes)) or (z > np.amax(planes))):
                 return np.array([])
@@ -715,70 +823,173 @@ class DicomParser:
                 # Fractional distance of dose plane between upper & lower bound
                 fz = (z - planes[lb]) / (planes[ub] - planes[lb])
                 plane = self.InterpolateDosePlanes(
-                    self.ds.pixel_array[ub], self.ds.pixel_array[lb], fz)
+                    pixel_array[ub], pixel_array[lb], fz)
                 return plane
         else:
             return np.array([])
 
     def InterpolateDosePlanes(self, uplane, lplane, fz):
-        """Interpolates a dose plane between two bounding planes at the given
-           relative location.
+        """Interpolate a dose plane between two bounding planes.
 
-        :param uplane:      Upper dose plane boundary.
-        :param uplane:      Lower dose plane boundary.
-        :param fz:          Fractional distance from the bottom to the top,
-                            where the new plane is located.
-                            E.g. if fz = 1, the plane is at the upper plane,
-                            fz = 0, it is at the lower plane.
-        :return:            An numpy 2d array of the interpolated dose plane.
+        Parameters
+        ----------
+        uplane : float
+            Upper dose plane boundary in mm
+        lplane : float
+            Lower dose plane boundary in mm
+        fz : float
+            Fractional distance from the bottom to the top,
+            where the new plane is located.
+            E.g. if fz = 1, the plane is at the upper plane,
+            fz = 0, it is at the lower plane.
+
+        Returns
+        -------
+        numpy array
+            An numpy 2d array of the interpolated dose plane
         """
-
         # A simple linear interpolation
         doseplane = fz*uplane + (1.0 - fz)*lplane
 
         return doseplane
 
-    def GetIsodosePoints(self, z=0, level=100, threshold=0.5):
-        """Return points for the given isodose level and slice position
-           from the dose grid.
+    def is_head_first_orientation(self):
+        """Return True if self.orientation is head-first.
+
+        Raises
+        ------
+        NotImplementedError
+            Raised if orientation is not one of head/feet-first
+            and supine/prone/decubitus
+
+        Returns
+        -------
+        bool
+            True if orientation is head-first, else False
+
 
-        :param z:           Slice position in mm.
-        :param threshold:   Threshold in mm to determine the max difference
-                            from z to the closest dose slice without
-                            using interpolation.
-        :param level:       Isodose level in scaled form
-                            (multiplied by self.ds.DoseGridScaling)
-        :return:            An array of tuples representing isodose points.
         """
+        orientation = self.ds.ImageOrientationPatient
+        if any(
+            all(np.isclose(orientation, hf_orientation))
+            for hf_orientation in (  # noqa
+                [1,  0,  0,  0,  1,  0],  # Head First Supine
+                [-1,  0,  0,  0, -1,  0],  # Head First Prone
+                [0, -1,  0,  1,  0,  0],  # Head First Decubitus Left
+                [0,  1,  0, -1,  0,  0]   # Head First Decubitus Right
+            )
+        ):
+            return True
+        elif any(
+            all(np.isclose(orientation, ff_orientation))
+            for ff_orientation in (
+                [0,  1,  0,  1,  0,  0],  # Feet First Decubitus Left
+                [0, -1,  0, -1,  0,  0],  # Feet First Decubitus Right
+                [1,  0,  0,  0, -1,  0],  # Feet First Prone
+                [-1,  0,  0,  0,  1,  0]   # Feet First Supine
+            )
+        ):
+            return False
+        else:
+            raise NotImplementedError(
+                "Cannot calculate dose plane sign for non-standard orientation"
+            )
 
+    def x_lut_index(self):
+        """Return LUT index for real-world X direction.
+
+        Raises
+        ------
+        NotImplementedError
+            Raised if orientation is not one of head/feet-first
+            and supine/prone/decubitus
+
+        Returns
+        -------
+        X direction LUT index, matching 'lut' from GetDoseData
+            0 if real-world X across columns
+            1 if real-world X along rows
+        """
+        orientation = self.ds.ImageOrientationPatient
+        if any(
+            all(np.isclose(orientation, non_decub))
+            for non_decub in (
+                [1,  0,  0,  0,  1,  0],  # Head First Supine
+                [-1,  0,  0,  0, -1,  0],  # Head First Prone
+                [-1,  0,  0,  0,  1,  0],  # Feet First Supine
+                [1,  0,  0,  0, -1,  0]   # Feet First Prone
+            )
+        ):
+            return 0
+        elif any(
+            all(np.isclose(orientation, decub))
+            for decub in (
+                [0, -1,  0,  1,  0,  0],  # Head First Decubitus Left
+                [0,  1,  0, -1,  0,  0],  # Head First Decubitus Right
+                [0,  1,  0,  1,  0,  0],  # Feet First Decubitus Left
+                [0, -1,  0, -1,  0,  0]   # Feet First Decubitus Right
+            )
+        ):
+            return 1
+        else:
+            raise NotImplementedError(
+                "Cannot calculate X direction for non-standard orientation"
+            )
+
+    def GetIsodosePoints(self, z=0, level=100, threshold=0.5):
+        """Return dose grid points from an isodose level & slice position.
+
+        Parameters
+        ----------
+        z : int, optional
+            Slice position in mm., by default 0
+        level : int, optional
+            Isodose level in scaled form
+            (multiplied by self.ds.DoseGridScaling), by default 100
+        threshold : float, optional
+            Threshold in mm to determine the max difference
+            from z to the closest dose slice without
+            using interpolation, by default 0.5
+
+        Returns
+        -------
+        list
+            An list of tuples representing isodose points
+        """
         plane = self.GetDoseGrid(z, threshold)
         isodose = (plane >= level).nonzero()
         return list(zip(isodose[1].tolist(), isodose[0].tolist()))
 
     def GetDoseData(self):
         """Return the dose data from a DICOM RT Dose file."""
-
         data = self.GetImageData()
-        data['doseunits'] = self.ds.DoseUnits
-        data['dosetype'] = self.ds.DoseType
-        data['dosecomment'] = self.ds.DoseComment \
-            if 'DoseComment' in self.ds else ''
-        data['dosesummationtype'] = self.ds.DoseSummationType
-        data['dosegridscaling'] = self.ds.DoseGridScaling
-        data['dosemax'] = float(self.ds.pixel_array.max())
+        data['doseunits'] = getattr(self.ds, 'DoseUnits', '')
+        data['dosetype'] = getattr(self.ds, 'DoseType', '')
+        data['dosecomment'] = getattr(self.ds, 'DoseComment', '')
+        data['dosesummationtype'] = getattr(self.ds, 'DoseSummationType', '')
+        data['dosegridscaling'] = getattr(self.ds, 'DoseGridScaling', '')
+        dosemax = 0
+        for x in range(data["frames"]):
+            pixel_array = self.GetPixelArray()
+            newmax = pixel_array[x].max()
+            dosemax = newmax if newmax > dosemax else dosemax
+            if self.memmap_pixel_array:
+                del pixel_array
+        data['dosemax'] = float(dosemax)
         data['lut'] = self.GetPatientToPixelLUT()
+        data['x_lut_index'] = self.x_lut_index()
         data['fraction'] = ''
         if "ReferencedRTPlanSequence" in self.ds:
             plan = self.ds.ReferencedRTPlanSequence[0]
             if "ReferencedFractionGroupSequence" in plan:
                 data['fraction'] = \
-                plan.ReferencedFractionGroupSequence[0].ReferencedFractionGroupNumber
+                    plan.ReferencedFractionGroupSequence[
+                        0].ReferencedFractionGroupNumber
 
         return data
 
     def GetReferencedBeamNumber(self):
         """Return the referenced beam number (if it exists) from RT Dose."""
-
         beam = None
         if "ReferencedRTPlanSequence" in self.ds:
             rp = self.ds.ReferencedRTPlanSequence[0]
@@ -791,16 +1002,15 @@ class DicomParser:
 
         return beam
 
-############################## RT Plan Methods ###############################
+# ============================== RT Plan Methods ==============================
 
     def GetPlan(self):
-        """Returns the plan information."""
-
+        """Return the plan information."""
         self.plan = {}
 
-        self.plan['label'] = self.ds.RTPlanLabel
-        self.plan['date'] = self.ds.RTPlanDate
-        self.plan['time'] = self.ds.RTPlanTime
+        self.plan['label'] = getattr(self.ds, 'RTPlanLabel', '')
+        self.plan['date'] = getattr(self.ds, 'RTPlanDate', '')
+        self.plan['time'] = getattr(self.ds, 'RTPlanTime', '')
         self.plan['name'] = ''
         self.plan['rxdose'] = 0
         if "DoseReferenceSequence" in self.ds:
@@ -816,7 +1026,8 @@ class DicomParser:
                 elif item.DoseReferenceStructureType == 'VOLUME':
                     if 'TargetPrescriptionDose' in item:
                         self.plan['rxdose'] = item.TargetPrescriptionDose * 100
-        if (("FractionGroupSequence" in self.ds) and (self.plan['rxdose'] == 0)):
+        if (("FractionGroupSequence" in self.ds) and
+                (self.plan['rxdose'] == 0)):
             fg = self.ds.FractionGroupSequence[0]
             if ("ReferencedBeamSequence" in fg) and \
                ("NumberOfFractionsPlanned" in fg):
@@ -833,8 +1044,18 @@ class DicomParser:
         return self.plan
 
     def GetReferencedBeamsInFraction(self, fx=0):
-        """Return the referenced beams from the specified fraction."""
+        """Return the referenced beams from the specified fraction.
 
+        Parameters
+        ----------
+        fx : int, optional
+            FractionGroupSequence number, by default 0
+
+        Returns
+        -------
+        dict
+            Dictionary of referenced beam data
+        """
         beams = {}
         if ("BeamSequence" in self.ds):
             bdict = self.ds.BeamSequence
