@@ -10,7 +10,7 @@
 import numpy as np
 import re
 import logging
-logger = logging.getLogger('dicompyler.dvh')
+logger = logging.getLogger('dicompylercore.dvh')
 
 # Set default absolute dose and volume  units
 abs_dose_units = 'Gy'
@@ -61,7 +61,7 @@ class DVH(object):
 
     @classmethod
     def from_dicom_dvh(cls, dataset, roi_num, rx_dose=None, name=None,
-            color=None):
+                       color=None):
         """Initialization for a DVH from a pydicom RT Dose DVH sequence."""
         sequence_num = -1
         for i, d in enumerate(dataset.DVHSequence):
@@ -76,8 +76,8 @@ class DVH(object):
                 "'DVHSequence' has no DVH with ROI Number '%d'." % roi_num)
         dvh = dataset.DVHSequence[sequence_num]
         data = np.array(dvh.DVHData)
-        return cls(counts=data[1::2] * dvh.DVHDoseScaling,
-                   bins=data[0::2].cumsum(),
+        return cls(counts=data[1::2],
+                   bins=data[0::2].cumsum() * dvh.DVHDoseScaling,
                    dvh_type=dvh.DVHType.lower(),
                    dose_units=dvh.DoseUnits.capitalize(),
                    volume_units=dvh.DVHVolumeUnits.lower(),
@@ -469,7 +469,7 @@ class DVH(object):
         if volume == 100 and not volume_units:
             # Flipping the difference volume array
             reversed_difference_of_volume = np.flip(
-                    np.fabs(volume_counts - volume), 0)
+                np.fabs(volume_counts - volume), 0)
 
             # Index of the first minimum value in reversed array
             index_min_value = np.argmin(reversed_difference_of_volume)
