@@ -15,6 +15,10 @@ import matplotlib.path
 from dicompylercore import dvh
 from dicompylercore.config import skimage_available
 import collections
+try:
+    from collections.abc import Sequence
+except ImportError:
+    from collections import Sequence
 from six import iteritems
 import logging
 logger = logging.getLogger('dicompylercore.dvhcalc')
@@ -32,14 +36,15 @@ def get_dvh(structure,
             interpolation_resolution=None,
             interpolation_segments_between_planes=0,
             thickness=None,
+            memmap_rtdose=False,
             callback=None):
     """Calculate a cumulative DVH in Gy from a DICOM RT Structure Set & Dose.
 
     Parameters
     ----------
-    structure : pydicom Dataset
+    structure : pydicom Dataset or filename
         DICOM RT Structure Set used to determine the structure data.
-    dose : pydicom Dataset
+    dose : pydicom Dataset or filename
         DICOM RT Dose used to determine the dose grid.
     roi : int
         The ROI number used to uniquely identify the structure in the structure
@@ -51,28 +56,39 @@ def get_dvh(structure,
         dose grid.
     use_structure_extents : bool, optional
         Limit the DVH calculation to the in-plane structure boundaries.
-    interpolation_resolution : float, optional
-        Resolution in mm to interpolate the structure and dose data to.
+    interpolation_resolution : tuple or float, optional
+        Resolution in mm (row, col) to interpolate structure and dose data to.
+        If float is provided, original dose grid pixel spacing must be square.
     interpolation_segments_between_planes : integer, optional
         Number of segments to interpolate between structure slices.
     thickness : float, optional
         Structure thickness used to calculate volume of a voxel.
+    memmap_rtdose : bool, optional
+        Use memory mapping to access the pixel array of the DICOM RT Dose.
+        This reduces memory usage at the expense of increased calculation time.
     callback : function, optional
         A function that will be called at every iteration of the calculation.
+
+    Returns
+    -------
+    dvh.DVH
+        An instance of dvh.DVH in cumulative dose. This can be converted to
+        different formats using the attributes and properties of the DVH class.
     """
     from dicompylercore import dicomparser
+
     rtss = dicomparser.DicomParser(structure)
-    rtdose = dicomparser.DicomParser(dose)
+    rtdose = dicomparser.DicomParser(dose, memmap_pixel_array=memmap_rtdose)
     structures = rtss.GetStructures()
     s = structures[roi]
     s['planes'] = rtss.GetStructureCoordinates(roi)
     s['thickness'] = thickness if thickness else rtss.CalculatePlaneThickness(
         s['planes'])
 
-    calcdvh = calculate_dvh(s, rtdose, limit, calculate_full_volume,
-                            use_structure_extents, interpolation_resolution,
-                            interpolation_segments_between_planes,
-                            callback)
+    calcdvh = _calculate_dvh(s, rtdose, limit, calculate_full_volume,
+                             use_structure_extents, interpolation_resolution,
+                             interpolation_segments_between_planes,
+                             callback)
     return dvh.DVH(counts=calcdvh.histogram,
                    bins=(np.arange(0, 2) if (calcdvh.histogram.size == 1) else
                          np.arange(0, calcdvh.histogram.size + 1) / 100),
@@ -82,20 +98,21 @@ def get_dvh(structure,
                    name=s['name']).cumulative
 
 
-def calculate_dvh(structure,
-                  dose,
-                  limit=None,
-                  calculate_full_volume=True,
-                  use_structure_extents=False,
-                  interpolation_resolution=None,
-                  interpolation_segments_between_planes=0,
-                  callback=None):
-    """Calculate the differential DVH for the given structure and dose grid.
+def _calculate_dvh(structure,
+                   dose,
+                   limit=None,
+                   calculate_full_volume=True,
+                   use_structure_extents=False,
+                   interpolation_resolution=None,
+                   interpolation_segments_between_planes=0,
+                   callback=None):
+    """Calculate a differential DVH for the given structure and dose grid.
 
     Parameters
     ----------
     structure : dict
-        A structure (ROI) from an RT Structure Set parsed using DicomParser
+        A structure (ROI) from an RT Structure Set parsed using DicomParser.
+        The dictionary must include a `thickness` key with a thickness `float`.
     dose : DicomParser
         A DicomParser instance of an RT Dose
     limit : int, optional
@@ -105,12 +122,23 @@ def calculate_dvh(structure,
         dose grid.
     use_structure_extents : bool, optional
         Limit the DVH calculation to the in-plane structure boundaries.
-    interpolation_resolution : float, optional
-        Resolution in mm to interpolate the structure and dose data to.
+    interpolation_resolution : tuple or float, optional
+        Resolution in mm (row, col) to interpolate structure and dose data to.
+        If float is provided, original dose grid pixel spacing must be square.
     interpolation_segments_between_planes : integer, optional
         Number of segments to interpolate between structure slices.
     callback : function, optional
         A function that will be called at every iteration of the calculation.
+
+    Returns
+    -------
+    calcdvh: collections.namedtuple
+        A named tuple of notes and and histogram data.
+
+    Notes
+    -----
+    This is an internal function called by `get_dvh` and
+    should not be called directly.
     """
     planes = collections.OrderedDict(sorted(iteritems(structure['planes'])))
     calcdvh = collections.namedtuple('DVH', ['notes', 'histogram'])
@@ -119,7 +147,7 @@ def calculate_dvh(structure,
 
     # Create an empty array of bins to store the histogram in cGy
     # only if the structure has contour data or the dose grid exists
-    if ((len(planes)) and ("PixelData" in dose.ds)):
+    if ((len(planes)) and (hasattr(dose, 'pixel_array'))):
 
         # Get the dose and image data information
         dd = dose.GetDoseData()
@@ -143,7 +171,7 @@ def calculate_dvh(structure,
                     dgindexextents,
                     dgextents,
                     new_pixel_spacing=interpolation_resolution,
-                    min_pixel_spacing=id['pixelspacing'][0])
+                    min_pixel_spacing=id['pixelspacing'])
             dd['rows'] = dd['lut'][1].shape[0]
             dd['columns'] = dd['lut'][0].shape[0]
 
@@ -151,7 +179,10 @@ def calculate_dvh(structure,
         # Code taken from Stack Overflow Answer from Joe Kington:
         # https://stackoverflow.com/q/3654289/74123
         # Create vertex coordinates for each grid cell
-        x, y = np.meshgrid(np.array(dd['lut'][0]), np.array(dd['lut'][1]))
+        x_index = dd['x_lut_index']
+        x, y = np.meshgrid(
+            np.array(dd['lut'][x_index]), np.array(dd['lut'][1-x_index])
+        )
         x, y = x.flatten(), y.flatten()
         dosegridpoints = np.vstack((x, y)).T
 
@@ -201,8 +232,18 @@ def calculate_dvh(structure,
                 logger.warning('Dose plane not found for %s.' +
                                ' Using %s to calculate contour volume.',
                                z, origin_z)
+                # Create a dummy dose grid with the correct size.
+                # calculate_plane_histogram() and its methods provide the
+                # volume calc needed, but do so as part of DVH calc,
+                # which requires a dose grid
+                dummy_dose = dose.GetDoseGrid(origin_z)
+                if use_structure_extents:
+                    extents = dgindexextents
+                    dummy_dose = dummy_dose[
+                        extents[1]:extents[3], extents[0]:extents[2]
+                    ]
                 _, vol = calculate_plane_histogram(
-                    plane, dose.GetDoseGrid(origin_z), dosegridpoints, maxdose,
+                    plane, dummy_dose, dosegridpoints, maxdose,
                     dd, id, structure, hist)
                 planedata[z] = (np.array([0]), vol)
                 notes = 'Dose grid does not encompass every contour.' + \
@@ -237,7 +278,7 @@ def calculate_plane_histogram(plane, doseplane, dosegridpoints, maxdose, dd,
     # and boolean xor to remove holes
     for i, contour in enumerate(contours):
         m = get_contour_mask(dd, id, dosegridpoints, contour)
-        grid = np.logical_xor(m.astype(np.uint8), grid).astype(np.bool)
+        grid = np.logical_xor(m.astype(np.uint8), grid).astype(np.bool_)
 
     hist, vol = calculate_contour_dvh(grid, doseplane, maxdose, dd, id,
                                       structure)
@@ -261,7 +302,10 @@ def get_contour_mask(dd, id, dosegridpoints, contour):
     # return mask.reshape((len(doselut[1]), len(doselut[0])))
 
     grid = c.contains_points(dosegridpoints)
-    grid = grid.reshape((len(doselut[1]), len(doselut[0])))
+    if dd['x_lut_index'] == 0:  # X values across columns
+        grid = grid.reshape((len(doselut[1]), len(doselut[0])))
+    else:  # decubitus
+        grid = grid.reshape((len(doselut[0]), len(doselut[1]))).T
 
     return grid
 
@@ -276,8 +320,8 @@ def calculate_contour_dvh(mask, doseplane, maxdose, dd, id, structure):
                                range=(0, maxdose))
 
     # Calculate the volume for the contour for the given dose plane
-    vol = sum(hist) * ((np.mean(np.diff(dd['lut'][0]))) *
-                       (np.mean(np.diff(dd['lut'][1]))) *
+    vol = sum(hist) * (abs(np.mean(np.diff(dd['lut'][0]))) *
+                       abs(np.mean(np.diff(dd['lut'][1]))) *
                        (structure['thickness']))
     return hist, vol
 
@@ -314,7 +358,7 @@ def dosegrid_extents_indices(extents, dd, padding=1):
     Parameters
     ----------
     extents : list
-        Structure extents in patient coordintes: [xmin, ymin, xmax, ymax].
+        Structure extents in patient coordinates: [xmin, ymin, xmax, ymax].
         If an empty list, no structure extents will be used in the calculation.
     dd : dict
         Dose data from dicomparser.GetDoseData.
@@ -324,24 +368,41 @@ def dosegrid_extents_indices(extents, dd, padding=1):
     Returns
     -------
     list
-        Dose grid extents in pixel coordintes as array indices:
-        [xmin, ymin, xmax, ymax].
+        Dose grid extents in pixel coordinates as array indices:
+        [col_min, row_min, col_max, row_max].
     """
+    col_lut, row_lut = dd['lut']
+    num_cols = len(col_lut)
+    num_rows = len(row_lut)
     if not len(extents):
-        return [0, 0, dd['lut'][0].shape[0] - 1, dd['lut'][1].shape[0] - 1]
-    dgxmin = np.argmin(np.fabs(dd['lut'][0] - extents[0])) - padding
-    if dd['lut'][0][dgxmin] > extents[0]:
-        dgxmin -= 1
-    dgxmax = np.argmin(np.fabs(dd['lut'][0] - extents[2])) + padding
-    dgymin = np.argmin(np.fabs(dd['lut'][1] - extents[1])) - padding
-    dgymax = np.argmin(np.fabs(dd['lut'][1] - extents[3])) + padding
-    dgxmin = 0 if dgxmin < 0 else dgxmin
-    dgymin = 0 if dgymin < 0 else dgymin
-    if dgxmax == dd['lut'][0].shape[0]:
-        dgxmax = dd['lut'][0].shape[0] - 1
-    if dgymax == dd['lut'][1].shape[0]:
-        dgymax = dd['lut'][1].shape[0] - 1
-    return [dgxmin, dgymin, dgxmax, dgymax]
+        return [0, 0, num_cols - 1, num_rows - 1]
+
+    if dd['x_lut_index'] == 0:  # X is across rows
+        strx_col_min, strx_col_max = extents[0], extents[2]
+        strx_row_min, strx_row_max = extents[1], extents[3]
+    else:  # decubitus case, X down rows
+        strx_col_min, strx_col_max = extents[1], extents[3]
+        strx_row_min, strx_row_max = extents[0], extents[2]
+
+    dg_col_min = np.argmin(np.fabs(col_lut - strx_col_min))
+    dg_col_max = np.argmin(np.fabs(col_lut - strx_col_max))
+    if dg_col_min > dg_col_max:
+        dg_col_min, dg_col_max = dg_col_max, dg_col_min
+    if col_lut[dg_col_min] > strx_col_min:
+        dg_col_min -= 1
+
+    dg_row_min = np.argmin(np.fabs(row_lut - strx_row_min))
+    dg_row_max = np.argmin(np.fabs(row_lut - strx_row_max))
+    if dg_row_min > dg_row_max:
+        dg_row_min, dg_row_max = dg_row_max, dg_row_min
+
+    # Ensure indices within array limits regardless of padding
+    dg_col_min = max(0, dg_col_min-padding)
+    dg_row_min = max(0, dg_row_min-padding)
+    dg_col_max = min(num_cols - 1, dg_col_max+padding)
+    dg_row_max = min(num_rows - 1, dg_row_max+padding)
+
+    return [dg_col_min, dg_row_min, dg_col_max, dg_row_max]
 
 
 def dosegrid_extents_positions(extents, dd):
@@ -350,7 +411,8 @@ def dosegrid_extents_positions(extents, dd):
     Parameters
     ----------
     extents : list
-        Dose grid extents in pixel coordintes: [xmin, ymin, xmax, ymax].
+        Dose grid extents in pixel coordintes:
+        [col_pos_min, row_pos_min, col_pos_max, row_pos_max].
     dd : dict
         Dose data from dicomparser.GetDoseData.
 
@@ -376,16 +438,19 @@ def get_resampled_lut(index_extents,
     index_extents : list
         Dose grid extents as array indices.
     extents : list
-        Dose grid extents in patient coordinates.
-    new_pixel_spacing : float
-        New pixel spacing in mm
-    min_pixel_spacing : float
-        Minimum pixel spacing used to determine the new pixel spacing
+        Dose grid extents in patient coordinates:
+        [col_pos_min, row_pos_min, col_pos_max, row_pos_max].
+    new_pixel_spacing : tuple or float
+        New pixel spacing in mm (row, column).
+        If float is provided, original dose grid pixel spacing must be square.
+    min_pixel_spacing : tuple
+        Min pixel spacing used to determine new pixel spacing (row, column).
 
     Returns
     -------
     tuple
-        A tuple of lists (x, y) of patient to pixel coordinate mappings.
+        A tuple of lut lists (col lut, row lut) with the coordinates of the
+        dose grid in patient coordinates
 
     Raises
     ------
@@ -398,6 +463,10 @@ def get_resampled_lut(index_extents,
     The new pixel spacing must be a factor of the original (minimum) pixel
     spacing. For example if the original pixel spacing was ``3`` mm, the new
     pixel spacing should be: ``3 / (2^n)`` mm, where ``n`` is an integer.
+    This applies independently to both the row and column pixel spacing.
+
+    If a single float value is provided it will be applied to both row and
+    column. Additionally, the original dose grid pixel spacing must be square.
 
     Examples
     --------
@@ -405,20 +474,39 @@ def get_resampled_lut(index_extents,
     Derived via: ``(3 / 2^16) == 0.375``
 
     """
-    if (min_pixel_spacing % new_pixel_spacing != 0.0):
+    if not isinstance(new_pixel_spacing, Sequence):
+        if not (min_pixel_spacing[0] == min_pixel_spacing[1]):
+            raise AttributeError(
+                "New pixel spacing must be provided as a (row, column) tuple.")
+        else:
+            new_pixel_spacing = (new_pixel_spacing, new_pixel_spacing)
+    if (min_pixel_spacing[0] % new_pixel_spacing[0] != 0.0):
         raise AttributeError(
-            "New pixel spacing must be a factor of %g/(2^n),"
-            % min_pixel_spacing +
-            " where n is an integer. Value provided was %g."
-            % new_pixel_spacing)
-    sampling_rate = np.array([
-        abs(index_extents[0] - index_extents[2]),
-        abs(index_extents[1] - index_extents[3])
-    ])
-    xsamples, ysamples = sampling_rate * min_pixel_spacing / new_pixel_spacing
-    x = np.linspace(extents[0], extents[2], int(xsamples), dtype=np.float)
-    y = np.linspace(extents[1], extents[3], int(ysamples), dtype=np.float)
-    return x, y
+            "New row pixel spacing must be a factor of %s/(2^n),"
+            % min_pixel_spacing[0] +
+            " where n is an integer. Value provided was %s."
+            % new_pixel_spacing[0])
+    if (min_pixel_spacing[1] % new_pixel_spacing[1] != 0.0):
+        raise AttributeError(
+            "New column pixel spacing must be a factor of %s/(2^n),"
+            % min_pixel_spacing[1] +
+            " where n is an integer. Value provided was %s."
+            % new_pixel_spacing[1])
+
+    # Existing number of cols, rows
+    num_cols = abs(index_extents[0] - index_extents[2])
+    num_rows = abs(index_extents[1] - index_extents[3])
+
+    col_samples = round(num_cols * min_pixel_spacing[1] / new_pixel_spacing[1])
+    row_samples = round(num_rows * min_pixel_spacing[0] / new_pixel_spacing[0])
+
+    col_lut = np.linspace(
+        extents[0], extents[2], int(col_samples), dtype=np.float64
+    )
+    row_lut = np.linspace(
+        extents[1], extents[3], int(row_samples), dtype=np.float64
+    )
+    return col_lut, row_lut
 
 
 def get_interpolated_dose(dose, z, resolution, extents):
@@ -430,8 +518,9 @@ def get_interpolated_dose(dose, z, resolution, extents):
         A DicomParser instance of an RT Dose.
     z : float
         Index in mm of z plane of dose grid.dose
-    resolution : float
+    resolution : tuple
         Interpolation resolution less than or equal to dose grid pixel spacing.
+        Provided in (row, col) format.
     extents : list
         Dose grid index extents.
 
@@ -442,6 +531,8 @@ def get_interpolated_dose(dose, z, resolution, extents):
     """
     # Return the dose bounded by extents if interpolation is not required
     d = dose.GetDoseGrid(z)
+    if not d.size:
+        return d  # cannot take 2d index below if empty
     extent_dose = d[extents[1]:extents[3],
                     extents[0]:extents[2]] if len(extents) else d
     if not resolution:
@@ -453,9 +544,11 @@ def get_interpolated_dose(dose, z, resolution, extents):
     interp_dose = rescale(
         extent_dose,
         scale=scale,
-        mode='symmetric',
         order=1,
-        preserve_range=True)
+        mode='symmetric',
+        preserve_range=True,
+        channel_axis=None
+    )
     return interp_dose
 
 
